from __future__ import annotations

import os
import sys
from pathlib import Path
from typing import Collection

import libcst as cst
from django.conf import ENVIRONMENT_VARIABLE as DJANGO_SETTINGS_MODULE_ENV_KEY
from libcst.codemod import CodemodContext

from .codemods import ForwardRelationTypingVisitor, RulesT, gather_codemods
from .django_utils import parse_models, setup_django
from .vendoring.monkeytype import MoveImportsToTypeCheckingBlockVisitor, get_newly_imported_items


def main(
    app_path: os.PathLike[str],
    settings_module: str | None = None,
    disabled_rules: Collection[RulesT] | None = None,
    type_checking_block: bool = True,
) -> None:
    settings_module = settings_module or os.getenv(DJANGO_SETTINGS_MODULE_ENV_KEY)
    disabled_rules = disabled_rules or []

    if not settings_module:
        raise ValueError("No value was provided for --settings-module, and no environment variable was set.")

    os.environ[DJANGO_SETTINGS_MODULE_ENV_KEY] = settings_module

    # Patching `sys.path` to allow Django to setup correctly
    sys.path.append(str(app_path))
    setup_django()

    codemods = gather_codemods(disabled_rules)
    model_infos = parse_models(Path(app_path))

    model_filenames = set(model_info.filename for model_info in model_infos)

    for filename in model_filenames:
        code = Path(filename).read_text(encoding="utf-8")
        source_module = cst.parse_module(code)

        context = CodemodContext(
            filename=filename,
            scratch={},
        )
        filename_models = [model for model in model_infos if Path(model.filename) == Path(filename)]

        # TODO use gathered codemods
        visitor = ForwardRelationTypingVisitor(
            context=context,
            model_infos=filename_models,
        )

        output_module = visitor.transform_module(source_module)

        if type_checking_block:
            newly_imported_items = get_newly_imported_items(output_module, source_module)
            context = CodemodContext()
            MoveImportsToTypeCheckingBlockVisitor.store_imports_in_context(
                context,
                newly_imported_items,
            )
            type_checking_block_transformer = MoveImportsToTypeCheckingBlockVisitor(context)
            output_module = type_checking_block_transformer.transform_module(output_module)
        print(output_module.code)



import sys
from pathlib import Path


def is_relative_to(path: Path, other: Path) -> bool:
    if sys.version_info >= (3, 9):
        return path.is_relative_to(other)
    else:
        try:
            path.relative_to(other)
            return True
        except ValueError:
            return False



from __future__ import annotations

from argparse import ArgumentParser, ArgumentTypeError, Namespace
from pathlib import Path

from .codemods import RulesT, rules
from .main import main


class ScriptNamespace(Namespace):
    path: Path
    """Path to the directory containing the Django application."""

    settings_module: str | None

    disable: list[RulesT] | None

    type_checking_block: bool


def _dir_path(path_str: str) -> Path:
    if not (path := Path(path_str)).is_dir():
        raise ArgumentTypeError(f"{path_str} must be an existing directory.")
    return path


def parse_args() -> ScriptNamespace:
    parser = ArgumentParser(
        "django-autotyping",
        "Add type hints to your models for better auto-completion.",
    )

    parser.add_argument(
        "path",
        type=_dir_path,
        help="Path to the directory containing the Django application. "
        "This directory should contain your `manage.py` file.",
    )
    parser.add_argument(
        "--settings-module",
        default=None,
        help="Value of the `DJANGO_SETTINGS_MODULE` environment variable (a dotted Python path).",
    )
    parser.add_argument(
        "--disable",
        choices=[rule[0] for rule in rules],
        nargs="*",
        help="Rules to be disabled.",
    )
    parser.add_argument(
        "--type-checking-block",
        action="store_true",
        default=False,
        help="Whether newly added imports should be in an `if TYPE_CHECKING` block (avoids circular imports).",
    )

    return parser.parse_args(namespace=ScriptNamespace())


def entrypoint() -> None:
    args = parse_args()

    main(args.path, args.settings_module, args.disable, args.type_checking_block)



from __future__ import annotations

import inspect
from dataclasses import dataclass, field
from types import ModuleType

from django.apps import AppConfig, apps
from django.db.models.fields.related import RelatedField

from .typing import ModelType


@dataclass
class ForwardRelation:
    class_name: str
    nullable: bool
    has_init_subclass: bool
    model: ModelType

    @property
    def model_module(self) -> ModuleType:
        """The module object of the model linked to this relation."""
        return inspect.getmodule(self.model)

    @property
    def app_config(self) -> AppConfig:
        """The `AppConfig` object where the model linked to this relation belongs."""
        return apps.get_app_config(self.model._meta.app_label)

    @property
    def app_models_module(self) -> ModuleType:
        """The module object where models of the app of the model linked to this relation are stored."""
        return self.app_config.models_module

    @classmethod
    def from_field(cls, field: RelatedField) -> ForwardRelation:
        return cls(
            class_name=field.__class__.__name__,
            nullable=field.null,
            has_init_subclass=hasattr(type(field), "__class_getitem__"),
            model=field.related_model,
        )


@dataclass
class ModelInfo:
    model: ModelType
    module: ModuleType
    forward_relations: dict[str, ForwardRelation] = field(default_factory=dict)

    @property
    def class_name(self) -> str:
        return self.model.__name__

    @property
    def filename(self) -> str:
        return inspect.getsourcefile(self.model)

    @property
    def app_config(self) -> AppConfig:
        """The `AppConfig` object where the model belongs."""
        return apps.get_app_config(self.model._meta.app_label)

    @classmethod
    def from_model(cls, model: ModelType) -> ModelInfo:
        forward_relations = {
            field.name: ForwardRelation.from_field(field)
            for field in model._meta.get_fields()
            if isinstance(field, RelatedField)  # TODO isinstance check on `Field`?
            # if field.many_to_one  # TODO may be unnecessary?
        }

        return cls(
            model=model,
            module=inspect.getmodule(model),
            forward_relations=forward_relations,
        )



from __future__ import annotations

import inspect
from pathlib import Path

import django
from django.apps import apps

from .compat import is_relative_to
from .models import ModelInfo
from .typing import ModelType


def setup_django() -> None:
    django.setup()


def _discover_models(root_dir: Path) -> list[ModelType]:
    """Discover models for all installed apps of the current project.

    Only the apps defined relative to `root_dir` will be returned.
    """
    return [model for model in apps.get_models() if is_relative_to(Path(inspect.getabsfile(model)), root_dir)]


def parse_models(root_dir: Path) -> list[ModelInfo]:
    return [ModelInfo.from_model(model) for model in _discover_models(root_dir.resolve())]



from __future__ import annotations

import libcst as cst
import libcst.matchers as m
from django.db.models.fields.related import RECURSIVE_RELATIONSHIP_CONSTANT
from libcst.codemod import CodemodContext, VisitorBasedCodemodCommand
from libcst.codemod.visitors import AddImportsVisitor
from libcst.metadata import ScopeProvider
from libcst.metadata.scope_provider import ClassScope

from ..models import ModelInfo

ASSIGN_FOREIGN_FIELD = m.Assign(
    value=m.Call(
        args=m.OneOf(
            (  # String reference as a kw: Field(to="Model"):
                m.ZeroOrMore(),
                m.Arg(value=m.SaveMatchedNode(m.SimpleString(), "string_reference"), keyword=m.Name("to")),
                m.ZeroOrMore(),
            ),
            (  # String reference as the first positional arg: Field("Model"):
                m.Arg(value=m.SaveMatchedNode(m.SimpleString(), "string_reference"), keyword=None),
                m.ZeroOrMore(),
            ),
        )
    )
)


class ForwardRelationTypingVisitor(VisitorBasedCodemodCommand):
    """A visitor that will add type annotations to forward relations.

    Rule identifier: `DJA001`.

    ```python
    from typing import TYPE_CHECKING

    from django.db import models

    # Model is imported in an `if TYPE_CHECKING` block if `--type-checking-block` is used.
    if TYPE_CHECKING:
        # Related model is imported from the corresponding apps models module:
        from myproject.reporters.models import Reporter


    class Article(models.Model):
        # If the field supports `__class_getitem__` at runtime, it is parametrized directly:
        reporter = models.ForeignKey["Reporter"](
            "reporters.Reporter",
            on_delete=models.CASCADE,
        )

        # Otherwise, an explicit annotation is used. No unnecessary import if model is in the same file.
        article_version: "models.OneToOneField[ArticleVersion]" = models.OneToOneField(
            "ArticleVersion",
            on_delete=models.CASCADE,
        )
    ```
    """

    METADATA_DEPENDENCIES = {ScopeProvider}

    def __init__(self, context: CodemodContext, model_infos: list[ModelInfo]) -> None:
        super().__init__(context)
        self.model_infos = model_infos
        self.current_model: ModelInfo | None = None

    def visit_ClassDef(self, node: cst.ClassDef) -> bool | None:
        scope = self.get_metadata(ScopeProvider, node)
        if type(scope) is ClassScope or not node.bases:  # TODO check for `object` base as well
            return False
        self.current_model = self.get_model_info(node)

    def leave_ClassDef(self, original_node: cst.ClassDef, updated_node: cst.ClassDef) -> cst.ClassDef:
        self.current_model = None
        return updated_node

    @m.leave(ASSIGN_FOREIGN_FIELD)
    def type_any_to_one_field(self, original_node: cst.Assign, updated_node: cst.Assign) -> cst.Assign | cst.AnnAssign:
        if self.current_model is None:
            return updated_node

        # TODO handle multiple targets? Unlikely in the context of a Django model
        target = updated_node.targets[0].target
        forward_relation = self.current_model.forward_relations.get(target.value)
        if forward_relation is None:
            return updated_node

        extracted = m.extract(updated_node, ASSIGN_FOREIGN_FIELD)
        string_reference: cst.SimpleString = extracted.get("string_reference")
        if not string_reference:
            return updated_node

        splitted = string_reference.raw_value.split(".")
        if len(splitted) == 1:
            # reference is in the same app
            class_ref = splitted[0]
        else:
            # reference is from an outside app, e.g. myapp.MyModel
            class_ref = splitted[1]
        if class_ref == RECURSIVE_RELATIONSHIP_CONSTANT:
            # Handle relationships with itself
            class_ref = self.current_model.class_name

        # TODO Run a transformer to move in an `if TYPE_CHECKING` block.
        if self.current_model.module is not forward_relation.model_module:
            # TODO check if model is from the same app: do relative import
            if self.current_model.app_config is forward_relation.app_config:
                pass
            else:
                AddImportsVisitor.add_needed_import(
                    self.context,
                    module=forward_relation.app_models_module.__name__,
                    obj=class_ref,
                )

        if forward_relation.has_init_subclass:
            # We can parametrize the field directly, we won't get runtime TypeErrors
            annotation_str = f'["{class_ref}"]'  # forward ref used here as it will be evaluated at runtime
            if isinstance(updated_node.value.func, cst.Name):
                # e.g. `field = ForeignKey(...)`
                updated_node.value.func.value += annotation_str
            else:
                # e.g. `field = models.ForeignKey(...)`
                updated_node.value.func.attr.value += annotation_str
            return updated_node
        else:
            # We explicitly annotate to avoid runtime TypeErrors
            # e.g. from `field = ForeignKey(...)` to `field: ForeignKey[...] = ForeignKey(...)`
            annotation_str = f"{_get_attribute_path(updated_node.value.func)}[{class_ref}]"
            return cst.AnnAssign(
                target=target,
                annotation=cst.Annotation(annotation=cst.SimpleString(value=f'"{annotation_str}"')),
                value=updated_node.value,
            )

    def get_model_info(self, node: cst.ClassDef) -> ModelInfo | None:
        # TODO use a provider instead?
        return next((model_info for model_info in self.model_infos if model_info.class_name == node.name.value), None)


def _get_attribute_path(node: cst.Name | cst.Attribute) -> str:
    """Get the dotted path to an object from the `cst.Call.func` attribute.

    `node` can either be a `cst.Name` or a `cst.Attribute` node, meaning both use cases are supported:

    ```py
    field_1 = ForeignKey(...)  # Returns 'ForeignKey'
    field_2 = models.ForeignKey(...)  # Returns 'models.ForeignKey'
    ```
    """
    if isinstance(node, cst.Name):
        return node.value

    if isinstance(node.value, cst.Attribute):
        prefix = _get_attribute_path(node.value)
        print("prefix", prefix)
    else:
        prefix = node.value.value
    return f"{prefix}.{node.attr.value}"



from __future__ import annotations

from typing import Literal

from libcst.codemod import VisitorBasedCodemodCommand

from .forward_relation_typing_visitor import ForwardRelationTypingVisitor

__all__ = ("ForwardRelationTypingVisitor", "rules", "gather_codemods")

rules = [
    ("DJA001", ForwardRelationTypingVisitor),
]

RulesT = Literal["DJA001"]


def gather_codemods(disabled: list[str]) -> list[type[VisitorBasedCodemodCommand]]:
    return [rule[1] for rule in rules if rule[0] not in disabled]



# Copyright (c) 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
from __future__ import annotations

from typing import cast

import libcst
from libcst import (
    BaseCompoundStatement,
    BaseSmallStatement,
    BaseSuite,
    CSTTransformer,
    FlattenSentinel,
    Import,
    ImportFrom,
    ImportStar,
    MaybeSentinel,
    Module,
    RemovalSentinel,
    RemoveFromParent,
    SimpleStatementLine,
)
from libcst.codemod import CodemodContext, ContextAwareTransformer
from libcst.codemod.visitors import AddImportsVisitor, GatherImportsVisitor, ImportItem
from libcst.helpers import get_absolute_module_from_package_for_import


class MoveImportsToTypeCheckingBlockVisitor(ContextAwareTransformer):
    CONTEXT_KEY = "MoveImportsToTypeCheckingBlockVisitor"

    def __init__(
        self,
        context: CodemodContext,
    ) -> None:
        super().__init__(context)

        self.import_items_to_be_moved: list[ImportItem] = []

    @staticmethod
    def store_imports_in_context(
        context: CodemodContext,
        import_items_to_be_moved: list[ImportItem],
    ) -> None:
        context.scratch[MoveImportsToTypeCheckingBlockVisitor.CONTEXT_KEY] = (import_items_to_be_moved,)

    @staticmethod
    def _add_type_checking_import(source_module: Module) -> Module:
        context = CodemodContext()
        AddImportsVisitor.add_needed_import(context, "typing", "TYPE_CHECKING")
        transformer = AddImportsVisitor(context)
        transformed_source_module = transformer.transform_module(source_module)
        return transformed_source_module

    def _remove_imports(self, tree: Module) -> Module:
        transformer = RemoveImportsTransformer(self.import_items_to_be_moved)
        transformed_source_module = tree.visit(transformer)
        return transformed_source_module

    def _get_import_module(self) -> Module:
        empty_code = libcst.parse_module("")
        context = CodemodContext()
        context.scratch[AddImportsVisitor.CONTEXT_KEY] = self.import_items_to_be_moved
        transformer = AddImportsVisitor(context)
        transformed_source_module = transformer.transform_module(empty_code)
        return transformed_source_module

    @staticmethod
    def _replace_pass_with_imports(placeholder_module: Module, import_module: Module) -> Module:
        return placeholder_module.with_deep_changes(
            old_node=cast(BaseSuite, placeholder_module.body[0].body),
            body=import_module.body,
        )

    def _split_module(
        self, module: Module
    ) -> tuple[list[SimpleStatementLine | BaseCompoundStatement], list[SimpleStatementLine | BaseCompoundStatement],]:
        type_checking_block_add_location = 0
        gatherer = GatherImportsVisitor(self.context)
        module.visit(gatherer)
        all_imports = gatherer.all_imports

        for i, statement in enumerate(module.body):
            if isinstance(statement, SimpleStatementLine):
                for possible_import in statement.body:
                    for last_import in all_imports:
                        if possible_import is last_import:
                            type_checking_block_add_location = i + 1
                            break

        return (
            list(module.body[:type_checking_block_add_location]),
            list(module.body[type_checking_block_add_location:]),
        )

    def _add_if_type_checking_block(self, module: Module) -> Module:
        if not self.import_items_to_be_moved:
            return module

        import_module = self._get_import_module()

        placeholder_module = libcst.parse_module("\nif TYPE_CHECKING:\n    pass\n")
        type_checking_block_module = self._replace_pass_with_imports(
            placeholder_module,
            import_module,
        )

        # Find the point of insertion for the TYPE_CHECKING block
        statements_before_imports, statements_after_imports = self._split_module(module)

        updated_body_list = [
            *statements_before_imports,
            type_checking_block_module,
            *statements_after_imports,
        ]

        return module.with_changes(body=updated_body_list)

    @staticmethod
    def _remove_typing_module(import_item_list: list[ImportItem]) -> list[ImportItem]:
        ret: list[ImportItem] = []
        for import_item in import_item_list:
            if import_item.module_name != "typing":
                ret.append(import_item)
        return ret

    def transform_module_impl(
        self,
        tree: Module,
    ) -> Module:
        # Add from typing import TYPE_CHECKING
        tree = self._add_type_checking_import(tree)

        context_contents = self.context.scratch.get(MoveImportsToTypeCheckingBlockVisitor.CONTEXT_KEY)
        if context_contents is not None:
            (import_items_to_be_moved,) = context_contents

            self.import_items_to_be_moved = import_items_to_be_moved

            # Remove typing library since we do not want it
            # to be imported inside the if TYPE_CHECKING block
            self.import_items_to_be_moved = self._remove_typing_module(self.import_items_to_be_moved)

            # Remove the newer imports since those are to be
            # shifted inside the if TYPE_CHECKING block
            tree = self._remove_imports(tree)

            # Add the new imports inside if TYPE_CHECKING block
            tree = self._add_if_type_checking_block(tree)

        return tree


class RemoveImportsTransformer(CSTTransformer):
    def __init__(
        self,
        import_items_to_be_removed: list[ImportItem],
    ) -> None:
        super().__init__()
        self.import_items_to_be_removed = import_items_to_be_removed

    def leave_Import(
        self, original_node: Import, updated_node: Import
    ) -> BaseSmallStatement | FlattenSentinel[BaseSmallStatement] | RemovalSentinel:
        names_to_keep = []
        for name in updated_node.names:
            module_name = name.evaluated_name
            found = False
            for import_item in self.import_items_to_be_removed:
                if import_item.module_name == module_name:
                    found = True
                    break
            if not found:
                names_to_keep.append(name.with_changes(comma=MaybeSentinel.DEFAULT))

        if not names_to_keep:
            return RemoveFromParent()
        else:
            return updated_node.with_changes(names=names_to_keep)

    def leave_ImportFrom(
        self, original_node: ImportFrom, updated_node: ImportFrom
    ) -> BaseSmallStatement | FlattenSentinel[BaseSmallStatement] | RemovalSentinel:
        if isinstance(updated_node.names, ImportStar):
            return updated_node

        names_to_keep = []
        module_name = get_absolute_module_from_package_for_import(None, updated_node)
        for name in updated_node.names:
            name_value = name.name.value
            found = False
            for import_item in self.import_items_to_be_removed:
                if import_item.module_name == module_name and import_item.obj_name == name_value:
                    found = True
                    break
            if not found:
                names_to_keep.append(name.with_changes(comma=MaybeSentinel.DEFAULT))

        if not names_to_keep:
            return RemoveFromParent()
        else:
            return updated_node.with_changes(names=names_to_keep)
